"""
Simplified implementation of AllenNLP Registrable:
https://github.com/allenai/allennlp
"""

from __future__ import annotations

import json
import logging
import os
from abc import ABC, abstractmethod
from collections import defaultdict
from copy import deepcopy
from typing import Any, Callable, ClassVar, DefaultDict, TypeVar, cast

from pydantic import (
    BaseModel,
    Field,
    ModelWrapValidatorHandler,
    model_serializer,
    model_validator,
)
from pydantic_core.core_schema import (
    SerializationInfo,
    SerializerFunctionWrapHandler,
    ValidatorFunctionWrapHandler,
)
from typing_extensions import Self

from icij_common.import_utils import VariableNotFound, import_variable
from icij_common.pydantic_utils import ICIJSettings, icij_config

logger = logging.getLogger(__name__)

_RegistrableT = TypeVar("_RegistrableT", bound="Registrable")
_SubclassRegistry = dict[str, _RegistrableT]
_SubclssNames = dict[_SubclassRegistry, str]

T = TypeVar("T", bound="FromConfig")


class RegistrableMixin(ABC):
    _registry: ClassVar[DefaultDict[type, _SubclassRegistry]] = defaultdict(dict)
    _registered_names: ClassVar[DefaultDict[type, _SubclssNames]] = defaultdict(dict)

    @classmethod
    def register(
        cls, name: str | None = None, exist_ok: bool = False
    ) -> Callable[[type[RegistrableMixin]], type[RegistrableMixin]]:
        # pylint: disable=protected-access
        registry = RegistrableMixin._registry[cls]
        registered_names = RegistrableMixin._registered_names[cls]

        def add_subclass_to_registry(
            subclass: type[RegistrableMixin],
        ) -> type[RegistrableMixin]:
            registered_name = name
            if registered_name is None:
                registered_key = subclass.registry_key.default
                if registered_key is None:
                    raise ValueError(
                        "no name provided and the class doesn't define a registry key"
                    )
                registered_name = getattr(subclass, registered_key).default

            if registered_name in registry:
                if exist_ok:
                    msg = (
                        f"{registered_name} has already been registered as "
                        f"{registry[registered_name].__name__}, but exist_ok=True, "
                        f"so overwriting with {cls.__name__}"
                    )
                    logger.info(msg)
                else:
                    msg = (
                        f"Cannot register {registered_name} as {cls.__name__}; "
                        f"name already in use for {registry[registered_name].__name__}"
                    )
                    raise ValueError(msg)
            registry[registered_name] = subclass
            registered_names[subclass] = registered_name
            return subclass

        return add_subclass_to_registry

    @classmethod
    def by_name(cls: type[_RegistrableT], name: str) -> Callable[..., _RegistrableT]:
        logger.debug("instantiating registered subclass %s of %s", name, cls)
        subclass = cls.resolve_class_name(name)
        return cast(type[_RegistrableT], subclass)

    @classmethod
    def resolve_class_name(cls: type[_RegistrableT], name: str) -> type[_RegistrableT]:
        # pylint: disable=protected-access
        sub_registry = RegistrableMixin._registry.get(cls, None)
        if sub_registry is None:
            for k, v in RegistrableMixin._registry.items():
                if issubclass(cls, k):
                    sub_registry = v
                    break
        if sub_registry is not None:
            subclass = sub_registry.get(name, None)
            if subclass is not None:
                return subclass
        if "." in name:
            try:
                subclass = import_variable(name)
            except ModuleNotFoundError as e:
                raise ValueError(
                    f"tried to interpret {name} as a path to a class "
                    f"but unable to import module {'.'.join(name.split('.')[:-1])}"
                ) from e
            except VariableNotFound as e:
                split = name.split(".")
                raise ValueError(
                    f"tried to interpret {name} as a path to a class "
                    f"but unable to find class {split[-1]} in {split[:-1]}"
                ) from e
            return subclass
        available = "\n-".join(cls.list_available())
        msg = f"""{name} is not a registered name for '{cls.__name__}'.
Available names are:
{available}

If your registered class comes from custom code, you'll need to import the\
 corresponding modules and use fully-qualified paths: "my_module.submodule.MyClass"
"""
        raise ValueError(msg)

    @classmethod
    def list_available(cls) -> list[str]:
        # pylint: disable=protected-access
        keys = list(RegistrableMixin._registry[cls].keys())
        return keys

    @classmethod
    @property
    def registered_name(cls) -> str:
        for (
            names
        ) in (
            RegistrableMixin._registered_names.values()
        ):  # pylint: disable=protected-access
            name = names.get(cls)
            if name is not None:
                return name
        raise ValueError("registration inconsistency")


class RegistrableConfig(BaseModel, RegistrableMixin):
    registry_key: ClassVar[str] = Field(frozen=True, default="name")

    @model_validator(mode="wrap")
    @classmethod
    def deserialize_with_registry_key(
        cls, value: Any, handler: ValidatorFunctionWrapHandler
    ) -> RegistrableConfig:
        if isinstance(value, dict):
            copied = deepcopy(value)
            registry_key = copied.get(cls.registry_key.default)
            if registry_key is None:
                return handler(copied)
            subcls = cls.resolve_class_name(registry_key)
            return subcls.model_validate(copied)
        return handler(value)


class RegistrableSettings(RegistrableConfig, ICIJSettings):
    @classmethod
    def from_env(cls):
        key = cls.registry_key.default
        prefix = cls.model_config["env_prefix"]
        if prefix is not None:
            key = prefix + key
        registry_key = find_in_env(key, cls.model_config["case_sensitive"])
        subcls = cls.resolve_class_name(registry_key)
        return subcls()


class FromConfig(ABC):
    @classmethod
    @abstractmethod
    def _from_config(cls, config: RegistrableConfig, **extras) -> FromConfig: ...


class RegistrableFromConfig(RegistrableMixin, FromConfig, ABC):
    @classmethod
    def from_config(cls, config: RegistrableConfig, **extras) -> Self:
        name = getattr(config, config.registry_key.default).default
        subcls = cls.resolve_class_name(name)
        return subcls._from_config(config, **extras)  # pylint: disable=protected-access


def find_variable_loc_in_env(variable: str, case_sensitive: bool) -> tuple[str, str]:
    if case_sensitive:
        try:
            return variable, os.environ[variable]
        except KeyError as e:
            raise ValueError(f"couldn't find {variable} in env variables") from e
    lowercase = variable.lower()
    for k, v in os.environ.items():
        if k.lower() == lowercase:
            return k, v
    raise ValueError(f"couldn't find {variable.upper()} in env variables")


def find_in_env(variable: str, case_sensitive: bool) -> str:
    return find_variable_loc_in_env(variable, case_sensitive)[1]


class Registrable(BaseModel, RegistrableMixin, ABC):
    model_config = icij_config()

    registry_key: ClassVar[str] = Field(frozen=True, default="@type")

    @classmethod
    def model_validate_json(
        cls,
        json_data: str | bytes | bytearray,
        *,
        strict: bool | None = None,
        context: Any | None = None,
    ) -> Self:
        if isinstance(json_data, bytes):
            data = json.loads(json_data.decode("utf-8"))
        elif isinstance(json_data, str):
            data = json.loads(json_data)
        elif isinstance(json_data, bytearray):
            data = json.loads(str(json_data, encoding="utf-8"))
        else:
            raise TypeError(
                f"unsupported type {type(json_data)}, expected bytes, str or bytearray"
            )
        key = data[cls.registry_key.default]
        subcls = cls.resolve_class_name(key)
        # TODO: try to use model_validate_json instead
        return subcls.model_validate(data, strict=strict, context=context)

    @model_serializer(mode="wrap")
    def serialize_with_registry_key(
        self, nxt: SerializerFunctionWrapHandler, info: SerializationInfo
    ) -> RegistrableConfig:
        serialized = nxt(self)
        include_key = bool(info.include) and self.registry_key.default in info.include
        include_key = include_key or not (
            bool(info.exclude) and self.registry_key.default in info.exclude
        )
        if include_key:
            serialized[self.registry_key.default] = self.registered_name
        return serialized

    @model_validator(mode="wrap")
    @classmethod
    def deserialize_with_registry_key(
        cls, value: Any, handler: ModelWrapValidatorHandler[Self]
    ) -> Registrable:
        if isinstance(value, dict):
            copied = deepcopy(value)
            registry_key = copied.pop(cls.registry_key.default, None)
            if registry_key is None:
                return handler(copied)
            subcls = cls.resolve_class_name(registry_key)
            return subcls.model_validate(copied)
        return handler(value)



# pylint: disable=redefined-outer-name
import contextlib
from typing import AsyncGenerator

import neo4j
import pytest
from neo4j import AsyncGraphDatabase

import icij_common
from icij_common.neo4j_.db import NEO4J_COMMUNITY_DB

NEO4J_TEST_PORT = 7688
NEO4J_TEST_USER = "neo4j"
NEO4J_TEST_PASSWORD = "theneo4jpassword"
NEO4J_TEST_AUTH = neo4j.basic_auth(NEO4J_TEST_USER, NEO4J_TEST_PASSWORD)


async def wipe_db(session: neo4j.AsyncSession):
    # Indices and constraints
    query = "CALL apoc.schema.assert({}, {})"
    await session.run(query)
    # Documents
    query = """MATCH (n)
DETACH DELETE n
    """
    await session.run(query)


@contextlib.asynccontextmanager
async def _build_neo4j_driver():
    uri = f"neo4j://127.0.0.1:{NEO4J_TEST_PORT}"
    async with AsyncGraphDatabase.driver(  # pylint: disable=not-async-context-manager
        uri, auth=NEO4J_TEST_AUTH
    ) as driver:
        yield driver


@pytest.fixture(scope="module")
async def neo4j_test_driver_module() -> AsyncGenerator[neo4j.AsyncDriver, None]:
    async with _build_neo4j_driver() as driver:
        async with driver.session(database=neo4j.DEFAULT_DATABASE) as sess:
            await wipe_db(sess)
        yield driver


@pytest.fixture(scope="session")
async def neo4j_test_driver_session() -> AsyncGenerator[neo4j.AsyncDriver, None]:
    async with _build_neo4j_driver() as driver:
        async with driver.session(database=neo4j.DEFAULT_DATABASE) as sess:
            await wipe_db(sess)
        yield driver


@pytest.fixture()
async def neo4j_test_driver() -> AsyncGenerator[neo4j.AsyncDriver, None]:
    async with _build_neo4j_driver() as driver:
        async with driver.session(database="neo4j") as sess:
            await wipe_db(sess)
        yield driver


@pytest.fixture(scope="session")
async def neo4j_test_session_session(
    neo4j_test_driver_session: neo4j.Driver,
) -> AsyncGenerator[neo4j.AsyncSession, None]:
    driver = neo4j_test_driver_session
    async with driver.session(database="neo4j") as sess:
        await wipe_db(sess)
        yield sess


@pytest.fixture(scope="module")
async def neo4j_test_session_module(
    neo4j_test_session_session: neo4j.AsyncSession,
) -> neo4j.AsyncSession:
    session = neo4j_test_session_session
    await wipe_db(session)
    return session


@pytest.fixture()
async def neo4j_test_session(
    neo4j_test_session_session: neo4j.AsyncSession,
) -> neo4j.AsyncSession:
    session = neo4j_test_session_session
    await wipe_db(session)
    return session


async def mocked_is_enterprise(_: neo4j.AsyncDriver) -> bool:
    return True


@contextlib.asynccontextmanager
async def _mocked_db_specific_session(
    neo4j_driver: neo4j.AsyncDriver, db: str  # pylint: disable=unused-argument
) -> neo4j.AsyncSession:
    async with neo4j_driver.session(database=NEO4J_COMMUNITY_DB) as sess:
        yield sess


async def _mocked_db_registry_db(
    neo4j_driver: neo4j.AsyncDriver,  # pylint: disable=unused-argument
) -> str:
    return NEO4J_COMMUNITY_DB


@pytest.fixture()
def mock_enterprise(monkeypatch):
    mock_enterprise_(monkeypatch)


def mock_enterprise_(monkeypatch):
    monkeypatch.setattr(
        icij_common.neo4j_.db, "db_specific_session", _mocked_db_specific_session
    )
    monkeypatch.setattr(
        icij_common.neo4j_.migrate, "db_specific_session", _mocked_db_specific_session
    )
    monkeypatch.setattr(
        icij_common.neo4j_.db, "database_registry_db", _mocked_db_registry_db
    )
    monkeypatch.setattr(icij_common.neo4j_.db, "is_enterprise", mocked_is_enterprise)



import importlib
from typing import Any


class VariableNotFound(ImportError):
    pass


def import_variable(name: str) -> Any:
    parts = name.split(".")
    submodule = ".".join(parts[:-1])
    variable_name = parts[-1]
    try:
        module = importlib.import_module(submodule)
    except ModuleNotFoundError as e:
        raise VariableNotFound(e.msg) from e
    try:
        subclass = getattr(module, variable_name)
    except AttributeError as e:
        raise VariableNotFound(e) from e
    return subclass



from collections import defaultdict
from datetime import datetime
from functools import cached_property
from typing import Annotated, Any, Callable, Mapping, cast

from pydantic import BaseModel, ConfigDict, PlainSerializer
from pydantic.fields import FieldInfo
from pydantic_settings import BaseSettings

SetIntStr = set[int | str]
DictIntStrAny = dict[int | str, Any]


def to_lower_camel(field: str) -> str:
    return "".join(
        w.capitalize() if i > 0 else w for i, w in enumerate(field.split("_"))
    )


def to_lower_snake_case(s: str):
    snake = "".join("_" + c.lower() if c.isupper() else c for c in s)
    if snake.startswith("_"):
        snake = snake[1:]
    return snake


_SCHEMAS = dict()


def safe_copy(
    obj: BaseModel, update: Mapping[str, Any] | None = None, deep: bool = False
):
    if obj.__class__ not in _SCHEMAS:
        _SCHEMAS[obj.__class__] = dict()
        # Model.copy is always without alias
        _SCHEMAS[obj.__class__] = obj.model_json_schema(by_alias=False)
    schema = _SCHEMAS[obj.__class__]
    if update:
        for field in update.keys():
            props = schema.get("properties")
            if props is None:
                prop_key = schema["$ref"].replace("#/$defs/", "")
                props = schema["$defs"][prop_key]["properties"]
            if field not in props:
                msg = f'Unknown attribute "{field}" for {obj.__class__}'
                raise AttributeError(msg)

    return cast(obj.__class__, obj.model_copy(update=update, deep=deep))


def icij_config() -> ConfigDict:
    return ConfigDict(
        frozen=True,
        extra="forbid",
        populate_by_name=True,
        ignored_types=(cached_property,),
        use_enum_values=True,
        validate_default=True,
    )


class ICIJSettings(BaseSettings): ...  # pylint: disable=multiple-statements


def get_field_default_value(attr: FieldInfo):
    # Ugly work around pydantic v1 limitations on Field and default values
    if isinstance(attr, FieldInfo):
        if attr.default_factory is not None:
            return attr.default_factory()
        return attr.default
    return attr


def lowercamel_case_config() -> dict:
    return ConfigDict(alias_generator=to_lower_camel)


def ignore_extra_config() -> dict:
    return ConfigDict(extra="ignore")


def no_enum_config() -> dict:
    return ConfigDict(use_enum_values=False)


def merge_configs(*configs) -> ConfigDict:
    merged = dict()
    for c in configs:
        merged.update(c)
    return ConfigDict(**merged)


ISODatetime = Annotated[
    datetime,
    PlainSerializer(
        lambda x: x.isoformat(), return_type=datetime, when_used="json-unless-none"
    ),
]


def generate_encoders_by_class_tuples(
    type_encoder_map: dict[Any, Callable[[Any], Any]],
) -> dict[Callable[[Any], Any], tuple[Any, ...]]:
    encoders_by_class_tuples: dict[Callable[[Any], Any], tuple[Any, ...]] = defaultdict(
        tuple
    )
    for type_, encoder in type_encoder_map.items():
        encoders_by_class_tuples[encoder] += (type_,)
    return encoders_by_class_tuples



from __future__ import annotations

import asyncio
import logging
import time

from collections.abc import Coroutine
from datetime import datetime
from enum import Enum, unique
from inspect import signature
from typing import Any, Callable, Sequence, Annotated

import neo4j
from packaging.version import Version

from neo4j.exceptions import ConstraintError
from pydantic import BaseModel, BeforeValidator, ConfigDict, Field

from icij_common.pydantic_utils import merge_configs, no_enum_config
from icij_common.neo4j_.constants import (
    MIGRATION_COMPLETED,
    MIGRATION_DB,
    MIGRATION_LABEL,
    MIGRATION_NODE,
    MIGRATION_STARTED,
    MIGRATION_STATUS,
    MIGRATION_VERSION,
)
from .db import (
    Database,
    create_database_tx,
    create_db,
    databases_tx,
    db_specific_session,
    registry_db_session,
)

logger = logging.getLogger(__name__)

TransactionFn = Callable[[neo4j.AsyncTransaction], Coroutine]
ExplicitTransactionFn = Callable[[neo4j.Session], Coroutine]
MigrationFn = TransactionFn | ExplicitTransactionFn

_MIGRATION_TIMEOUT_MSG = """Migration timeout expired !
Please check that a migration is indeed in progress. If the application is in a \
deadlock restart it forcing the migration index cleanup."""


class MigrationError(RuntimeError):
    pass


@unique
class MigrationStatus(str, Enum):
    IN_PROGRESS = "IN_PROGRESS"
    DONE = "DONE"


def str_as_version(v: Any) -> MigrationVersion:
    if isinstance(v, (str, MigrationVersion)):
        return MigrationVersion(v)
    raise ValueError(
        f"Must be a {MigrationVersion.__name__} or a {str.__name__}, "
        f"found {type(v)}"
    )


MigrationVersion = Annotated[Version, BeforeValidator(str_as_version)]


class _BaseMigration(BaseModel):
    model_config = merge_configs(
        ConfigDict(arbitrary_types_allowed=True, populate_by_name=True),
        no_enum_config(),
    )

    version: MigrationVersion
    label: str
    status: MigrationStatus = MigrationStatus.IN_PROGRESS


class Neo4jMigration(_BaseMigration):
    # It would have been cleaner to create
    # (p:_Project)-[:_RUNS { id: p.name + m.version }]->(m:_Migration)
    # relationships. However, neo4j < 5.7 doesn't support unique constraint on
    # relationships properties which prevents from implementing the locking mechanism
    # properly. We hence enforce unique constraint on
    # (_Migration.version, _Migration.project)
    db: str = Field(alias="project")
    started: datetime
    completed: datetime | None = None
    status: MigrationStatus = MigrationStatus.IN_PROGRESS

    @classmethod
    def from_neo4j(cls, record: neo4j.Record, key="migration") -> Self:
        migration = dict(record.value(key))
        if "started" in migration:
            migration["started"] = migration["started"].to_native()
        if "completed" in migration:
            migration["completed"] = migration["completed"].to_native()
        return Neo4jMigration(**migration)


class Migration(_BaseMigration):
    migration_fn: MigrationFn


MigrationRegistry: Sequence[Migration]


async def _migrate_with_lock(
    *,
    db_session: neo4j.AsyncSession,
    registry_session: neo4j.AsyncSession,
    db: str,
    migration: Migration,
):
    # Note: all migrations.py should be carefully tested otherwise they will lock
    # the DB...

    # Lock the DB first, raising in case a migration already exists
    logger.debug("Trying to run migration to %s...", migration.label)
    await registry_session.execute_write(
        create_migration_tx,
        db=db,
        migration_version=str(migration.version),
        migration_label=migration.label,
    )
    # Then run to migration
    logger.debug("Acquired write lock to %s !", migration.label)
    sig = signature(migration.migration_fn)
    first_param = list(sig.parameters)[0]
    if first_param == "tx":
        await db_session.execute_write(migration.migration_fn)
    elif first_param == "sess":
        await migration.migration_fn(db_session)
    else:
        raise ValueError(f"Invalid migration function: {migration.migration_fn}")
    # Finally free the lock
    await registry_session.execute_write(
        complete_migration_tx,
        db=db,
        migration_version=str(migration.version),
    )
    logger.debug("Marked %s as complete !", migration.label)


async def create_migration_tx(
    tx: neo4j.AsyncTransaction,
    *,
    db: str,
    migration_version: str,
    migration_label: str,
) -> Neo4jMigration:
    query = f"""CREATE (m:{MIGRATION_NODE} {{
    {MIGRATION_DB}: $db,
    {MIGRATION_LABEL}: $label,
    {MIGRATION_VERSION}: $version,
    {MIGRATION_STATUS}: $status,
    {MIGRATION_STARTED}: $started
}})
RETURN m as migration"""
    res = await tx.run(
        query,
        label=migration_label,
        version=migration_version,
        db=db,
        status=MigrationStatus.IN_PROGRESS.value,
        started=datetime.now(),
    )
    migration = await res.single()
    if migration is None:
        raise ValueError(f"Couldn't find migration {migration_version} for {db}")
    migration = Neo4jMigration.from_neo4j(migration)
    return migration


async def complete_migration_tx(
    tx: neo4j.AsyncTransaction, *, db: str, migration_version: str
) -> Neo4jMigration:
    query = f"""MATCH (m:{MIGRATION_NODE} {{
        {MIGRATION_VERSION}: $version,
        {MIGRATION_DB}: $db
     }})
SET m += {{ {MIGRATION_STATUS}: $status, {MIGRATION_COMPLETED}: $completed }} 
RETURN m as migration"""
    res = await tx.run(
        query,
        version=migration_version,
        db=db,
        status=MigrationStatus.DONE.value,
        completed=datetime.now(),
    )
    migration = await res.single()
    migration = Neo4jMigration.from_neo4j(migration)
    return migration


async def db_migrations_tx(tx: neo4j.AsyncTransaction, db: str) -> list[Neo4jMigration]:
    query = f"""MATCH (m:{MIGRATION_NODE} {{ {MIGRATION_DB}: $db }})
RETURN m as migration
"""
    res = await tx.run(query, db=db)
    migrations = [Neo4jMigration.from_neo4j(rec) async for rec in res]
    return migrations


async def delete_all_migrations(driver: neo4j.AsyncDriver):
    query = f"""MATCH (m:{MIGRATION_NODE})
DETACH DELETE m"""
    async with registry_db_session(driver) as sess:
        await sess.run(query)


async def retrieve_dbs(neo4j_driver: neo4j.AsyncDriver) -> list[Database]:
    async with registry_db_session(neo4j_driver) as sess:
        dbs = await sess.execute_read(databases_tx)
    return dbs


async def migrate_db_schemas(
    neo4j_driver: neo4j.AsyncDriver,
    registry: MigrationRegistry,
    *,
    timeout_s: float,
    throttle_s: float,
):
    dbs = await retrieve_dbs(neo4j_driver)
    tasks = [
        migrate_db_schema(
            neo4j_driver,
            registry,
            db=p.name,
            timeout_s=timeout_s,
            throttle_s=throttle_s,
        )
        for p in dbs
    ]
    await asyncio.gather(*tasks)


async def migrate_db_schema(
    neo4j_driver: neo4j.AsyncDriver,
    registry: MigrationRegistry,
    db: str,
    *,
    timeout_s: float,
    throttle_s: float,
):
    logger.info("Migrating DB %s", db)
    start = time.monotonic()
    if not registry:
        return
    todo = sorted(registry, key=lambda m: m.version)
    async with registry_db_session(neo4j_driver) as registry_sess:
        async with db_specific_session(neo4j_driver, db=db) as db_session:
            while "Waiting for DB to be migrated or for a timeout":
                migrations = await registry_sess.execute_read(db_migrations_tx, db=db)
                in_progress = [
                    m for m in migrations if m.status is MigrationStatus.IN_PROGRESS
                ]
                if len(in_progress) > 1:
                    raise MigrationError(
                        f"Found several migration in progress: {in_progress}"
                    )
                if in_progress:
                    logger.info(
                        "Found that %s is in progress, waiting for %s seconds...",
                        in_progress[0].label,
                        throttle_s,
                    )
                    await asyncio.sleep(throttle_s)
                else:
                    done = [m for m in migrations if m.status is MigrationStatus.DONE]
                    if done:
                        current_version = max((m.version for m in done))
                        todo = [m for m in todo if m.version > current_version]
                    if not todo:
                        break
                    try:
                        await _migrate_with_lock(
                            db_session=db_session,
                            registry_session=registry_sess,
                            db=db,
                            migration=todo[0],
                        )
                        todo = todo[1:]
                        continue
                    except ConstraintError:
                        logger.info(
                            "Migration %s has just started somewhere else, "
                            " waiting for %s seconds...",
                            todo[0].label,
                            throttle_s,
                        )
                        await asyncio.sleep(throttle_s)
                elapsed = time.monotonic() - start
                if elapsed > timeout_s:
                    logger.error(_MIGRATION_TIMEOUT_MSG)
                    raise MigrationError(_MIGRATION_TIMEOUT_MSG)
                continue


async def init_database(
    neo4j_driver: neo4j.AsyncDriver,
    name: str,
    registry: MigrationRegistry,
    *,
    timeout_s: float,
    throttle_s: float,
) -> bool:
    # Create DB
    await create_db(neo4j_driver, db=name)

    # Record the DB in the registry
    async with registry_db_session(neo4j_driver) as sess:
        db, already_exists = await sess.execute_write(create_database_tx, name=name)

    # Migrate it
    await migrate_db_schema(
        neo4j_driver,
        registry=registry,
        db=db.name,
        timeout_s=timeout_s,
        throttle_s=throttle_s,
    )

    return already_exists



import logging
import traceback

from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.utils import is_body_allowed_for_status_code
from pydantic_core import ErrorDetails
from starlette import status
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse, Response

INTERNAL_SERVER_ERROR = "Internal Server Error"
_REQUEST_VALIDATION_ERROR = "Request Validation Error"

logger = logging.getLogger(__name__)


def json_error(*, title, detail, **kwargs) -> dict:
    error = {"title": title, "detail": detail}
    error.update(kwargs)
    return error


def display_errors(errors: list[ErrorDetails]) -> str:
    return "\n".join(
        f'{_display_error_loc(e)}\n  {e["msg"]} ({_display_error_type_and_ctx(e)})'
        for e in errors
    )


def _display_error_loc(error: ErrorDetails) -> str:
    return " -> ".join(str(e) for e in error["loc"])


def _display_error_type_and_ctx(error: ErrorDetails) -> str:
    t = "type=" + error["type"]
    ctx = error.get("ctx")
    if ctx:
        return t + "".join(f"; {k}={v}" for k, v in ctx.items())
    return t


async def request_validation_error_handler(
    request: Request, exc: RequestValidationError
):
    title = _REQUEST_VALIDATION_ERROR
    detail = display_errors(list(exc.errors()))
    error = json_error(title=title, detail=detail)
    logger.error("%s\nURL: %s\nDetail: %s", title, request.url, detail)
    return JSONResponse(
        status_code=status.HTTP_400_BAD_REQUEST, content=jsonable_encoder(error)
    )


async def http_exception_handler(request: Request, exc: StarletteHTTPException):
    headers = getattr(exc, "headers", None)
    if not is_body_allowed_for_status_code(exc.status_code):
        return Response(status_code=exc.status_code, headers=headers)
    title = detail = exc.detail
    error = json_error(title=title, detail=detail)
    logger.error("%s\nURL: %s", title, request.url)
    return JSONResponse(
        jsonable_encoder(error), status_code=exc.status_code, headers=headers
    )


async def internal_exception_handler(request: Request, exc: Exception):
    # pylint: disable=unused-argument
    title = INTERNAL_SERVER_ERROR
    detail = f"{type(exc).__name__}: {exc}"
    trace = "".join(traceback.format_exc())
    error = json_error(title=title, detail=detail, trace=trace)
    logger.exception(
        "%s\nURL: %s\nDetail: %s",
        title,
        request.url,
        detail,
    )
    return JSONResponse(jsonable_encoder(error), status_code=500)
